@@ -28,11 +28,45 @@ from async_lru import alru_cache
 from pandas import DataFrame, NamedAgg, Series
 from pydantic import Field
 
-from ..utils import LOG, HashableConfig
-from .tasks import EntityExtractor, SerpTopicAndIntentAssigner, SerpTopicExtractor
+from ..tools.flex.classify import Classifier
+from ..tools.flex.entities import EntityExtractor
+from ..tools.flex.topics import TopicAssigner, TopicExtractor
+from ..utils import LOG, Configurable, dedent
+
+TOPIC_INSTRUCTIONS = dedent("""
+Data records represent Google search keywords and associated SERP data.
+Make sure to create topics relevant in the context of SEO keywords research,
+focusing on the semantic meaning of keywords and SERPS, commercial intent etc.
+""")
+
+INTENT_INSTRUCTIONS = dedent("""
+Data records represent Google search keywords and associated SERP data.
+Categorize their search intent in the context of a SEO keywords analysis,
+focusing on the semantic meaning and intent.
+""")
+
+INTENTS = {
+    "informational": "The search intent is to find information or learn about a topic.",
+    "navigational": "The search intent is to navigate to a specific website or page.",
+    "transactional": "The search intent is to make a purchase or complete a transaction.",
+    "commercial": "The search intent is to compare products or services before making a decision.",
+    "local": "The search intent is to find local businesses or services in a specific area.",
+}
+
+ENTITY_INSTRUCTIONS = dedent("""
+Data records represent AI Overviews from SERP data associated with keyword searches.
+Focus on extracting entities that are relevant to SEO analysis.
+""")
+
+ENTITIES = {
+    "brand_company": "A brand or company name, typically a well-known entity in the market.",
+    "product_service": "A specific product or service offered by a brand or company.",
+    "technology": "A technology or platform related to the product or service.",
+    "other": "Any other relevant entity that does not fit into the above categories.",
+}
 
 
-class SerpConfig(HashableConfig):
+class SerpConfig(Configurable):
     """Configuration for fetching SERP data using Apify Google Search Scraper actor."""
 
     keywords: tuple[str, ...] | None = None
@@ -43,12 +77,18 @@ class SerpConfig(HashableConfig):
     """Number of results to fetch per page."""
     maxPagesPerQuery: int = 1
     """Maximum number of pages to fetch per query."""
-    country: str = "us"
-    """Country code for SERP data (e.g., 'us' for United States)."""
+    countryCode: str = "us"
+    """Specifies the country used for the search and the Google Search domain (e.g. google.es for
+    Spain). By default, the actor uses United States (google.com).
+    """
     searchLanguage: str = ""
-    """Search language for SERP data (e.g., 'en' for English)."""
+    """Restricts search results to pages in a specific language. For example, choosing 'German'
+    results in pages only in German. Passed to Google Search as the lr URL query parameter.
+    """
     languageCode: str = ""
-    """Language code for SERP data (e.g., 'en' for English)."""
+    """Language of the Google Search interface, not the search results themselves. Passed to
+    Google Search as the hl URL query parameter.
+    """
     params: dict[str, Any] | None = Field(default_factory=dict)
     """Additional parameters to pass to the Apify actor."""
     top_n: int = 10
@@ -61,6 +101,8 @@ class SerpConfig(HashableConfig):
     """Maximum number of samples to use for topic and intent extraction from SERP data."""
     topic_model: str | None = "google/gemini-2.5-flash-preview-05-20"
     """Model to use for topic extraction from SERP organic results."""
+    topic_min_ldist: int = 2
+    """Minimum Levenshtein distance between topic labels."""
     assignment_model: str | None = "openai/gpt-4.1-mini"
     """Model to use for intent classification from SERP organic results."""
     entity_model: str | None = "openai/gpt-4.1-mini"
@@ -99,7 +141,7 @@ async def fetch_serps(
     actor_param_names = (
         "resultsPerPage",
         "maxPagesPerQuery",
-        "country",
+        "countryCode",
         "searchLanguage",
         "languageCode",
     )
@@ -137,7 +179,6 @@ def process_toplevel_keys(row: dict):
         "#debug",
         "#error",
         "htmlSnapshotUrl",
-        "url",
         "hasNextPage",
         "resultsTotal",
         "serpProviderCode",
@@ -174,9 +215,15 @@ def process_ai_overview(row: dict):
     """Keep only content and source titles."""
     aio = row.pop("aiOverview", {})
     items = {
-        "aiOverview_content": aio.get("content", None),
-        "aiOverview_source_titles": [s["title"] for s in aio.get("sources", [])] or None,
+        "aio_content": aio.get("content", None),
+        "aio_source_titles": [s["title"] for s in aio.get("sources", [])] or None,
+        "aio_source_descs": [s["description"] for s in aio.get("sources", [])] or None,
+        "aio_source_urls": [s["url"] for s in aio.get("sources", [])] or None,
     }
+    items["aio_sources"] = [
+        s["title"] + " – " + s["description"] + " " + s["url"] for s in aio.get("sources", [])
+    ] or None
+
     row.update(**items)
 
 
@@ -288,16 +335,17 @@ def aggregate_organic_results(df: DataFrame, top_n=10) -> DataFrame:
     # These apply to only the top N results
     top_agg_funcs = {
         "titles": NamedAgg("title", list),
+        "urls": NamedAgg("url", list),
         "descriptions": NamedAgg("description", list),
         "domains": NamedAgg("domain", lambda ser: list(set(ser))),
         "breadcrumbs": NamedAgg("breadcrumb", lambda ser: list(set(flatten(ser)))),
         "emphasizedKeywords": NamedAgg("emphasizedKeywords", lambda ser: list(set(flatten(ser)))),
     }
 
-    agg = df.groupby("term").agg(**agg_funcs).reset_index()
+    agg = df.groupby("term").agg(**agg_funcs).reset_index()  # type: ignore[call-overload]
 
     top = df.groupby("term").head(top_n)
-    topagg = top.groupby("term").agg(**top_agg_funcs).reset_index()
+    topagg = top.groupby("term").agg(**top_agg_funcs).reset_index()  # type: ignore[call-overload]
 
     return agg.merge(topagg, on="term", how="left")
 
@@ -319,25 +367,27 @@ def add_ranks(
     df: DataFrame,
     brands: str | list[str] | None,
     competitors: str | list[str] | None,
+    columns: tuple[str, ...] | list[str] = ("titles", "domains", "descriptions"),
 ) -> DataFrame:
     """Calculate brand and competitor ranks in organic search results."""
+    missing = [col for col in columns if col not in df.columns]
+    if missing:
+        LOG.warning(f"Columns {missing} not found in DataFrame! Will not add ranks for these.")
+        columns = [col for col in columns if col not in missing]
+
     if brands is not None:
-        df["title_rank_brand"] = df.titles.apply(lambda x: token_rank(brands, x))
-        df["domain_rank_brand"] = df.domains.apply(lambda x: token_rank(brands, x))
-        df["description_rank_brand"] = df.descriptions.apply(lambda x: token_rank(brands, x))
+        for col in columns:
+            df[f"{col}_rank_brand"] = df[col].apply(lambda x: token_rank(brands, x))
 
     if competitors is not None:
         # First position of any(!) competitor
-        df["title_rank_competition"] = df.titles.apply(lambda x: token_rank(competitors, x))
-        df["description_rank_competition"] = df.descriptions.apply(
-            lambda x: token_rank(competitors, x)
-        )
-        df["domain_rank_competition"] = df.domains.apply(lambda x: token_rank(competitors, x))
+        for col in columns:
+            df[f"{col}_rank_competition"] = df[col].apply(lambda x: token_rank(competitors, x))
 
         # Specific ranks for each individual competitor
         for name in competitors:
             c_ranks = []
-            for col in ("titles", "descriptions", "domains"):
+            for col in columns:
                 rank = df[col].apply(lambda x, name=name: token_rank(name, x))
                 c_ranks.append(rank)
 
@@ -347,66 +397,100 @@ def add_ranks(
     return df
 
 
-async def topic_and_intent(
+async def topic_and_intent(  # noqa: PLR0913
     df: DataFrame,
     max_samples: int,
-    topic_model: str,
-    assignment_model: str,
+    topic_model: str = "google/gemini-2.5-flash-preview-05-20",
+    assignment_model: str = "openai/gpt-4.1-mini",
     max_retries: int = 5,
+    text_column: str = "term",
+    extra_columns: list[str] | None = None,
+    topics_instructions: str = "",
+    min_ldist: int = 2,
 ) -> DataFrame | None:
     """Classify keywords and their top N organic results into topics and intent."""
-    try:
-        extractor = SerpTopicExtractor(max_samples=max_samples, model=topic_model)  # type: ignore
-        topic_intent = await extractor(df, max_retries=max_retries)
-        LOG.info("Extracted topic hierarchy")
-        LOG.info(json.dumps(topic_intent.to_dict(), indent=2, ensure_ascii=False))
+    columns = [text_column] + (extra_columns or ["titles", "domains", "breadcrumbs"])
+    topic_instructions = TOPIC_INSTRUCTIONS + f"\n\n{topics_instructions}".strip()
 
-        assigner = SerpTopicAndIntentAssigner(topic_intent)
-        classified = await assigner(df=df, model=assignment_model, n_concurrent=100)
-        clf = classified.to_pandas()
-        return clf[["term", "topic", "subtopic", "intent"]]
+    try:
+        extractor = TopicExtractor(
+            model=topic_model,
+            records=df,
+            attrs=columns,
+            n_topics=10,
+            n_subtopics=5,
+            instructions=topic_instructions,
+            min_ldist=min_ldist,
+            max_samples=max_samples,
+        )
+        topics = await extractor(max_retries=max_retries)
+        LOG.info("Extracted topic hierarchy")
+        LOG.info(json.dumps(topics.to_dict(), indent=2, ensure_ascii=False))
+
+        assigner = TopicAssigner(
+            model=assignment_model,
+            records=df,
+            attrs=columns,
+            topics=topics,
+        )
+        term_topics = await assigner(n_concurrent=100)
+        term_topics = term_topics[[text_column, "topic", "subtopic"]]
+
+        classifier = Classifier(
+            model=assignment_model,
+            records=df,
+            attrs=columns,
+            categories=INTENTS,
+            instructions=INTENT_INSTRUCTIONS,
+        )
+        term_intents = await classifier(n_concurrent=100)
+        term_intents = term_intents[[text_column, "label"]]
+        term_intents = term_intents.rename(columns={"label": "intent"})
+
+        return term_topics.merge(term_intents, on=text_column, how="left")
     except Exception as exc:
         LOG.error(f"Error during topic and intent extraction: {exc}")
         LOG.exception("Stack trace:")
         return None
 
 
-async def process_ai_overviews(
+async def extract_aio_entities(
     df: DataFrame,
     entity_model: str = "openai/gpt-4.1-mini",
+    id_column: str = "term",
 ) -> DataFrame | None:
     """Process AI overviews in SERP data and extract entities."""
-    if "aiOverview_content" in df.columns and df["aiOverview_content"].notna().any():
-        try:
-            # Todo: extract brand and competitor ranks
-            ai_df = df[df.aiOverview_content.notna()].copy().reset_index()
-            entity_extractor = EntityExtractor()
-            entities = await entity_extractor(df=ai_df, model=entity_model, n_concurrent=100)
-            ent_df = entities.to_pandas(explode=False)
+    if ("aio_content" not in df.columns) or df["aio_content"].isna().all():
+        LOG.warning("No AI overview content found in DataFrame, skipping entity extraction.")
+        return None
 
-            for kind in ("brand/company", "product/service", "technology"):
-                ent_df[f"ai_overview_{kind}"] = ent_df.entities.apply(
-                    lambda es, kind=kind: [
-                        e.name
-                        for e in es
-                        if e is not None
-                        and hasattr(e, "type")
-                        and hasattr(e, "name")
-                        and e.type == kind
-                    ]
-                    if es is not None
-                    else None
-                )
+    aio_columns = [
+        "aio_content",
+        "aio_source_titles",
+        "aio_source_descs",
+        "aio_source_urls",
+    ]
+    aio_columns = [col for col in aio_columns if col in df.columns]
+    LOG.info(f"Extracting entities from columns: {aio_columns}")
 
-            ent_df["term"] = ai_df["term"]
-            return ent_df.drop(
-                columns=["aiOverview_content", "aiOverview_source_titles", "entities"]
-            )
-        except Exception as exc:
-            LOG.error(f"Error processing AI overviews: {exc}")
-            return None
+    try:
+        # Todo: extract brand and competitor ranks
+        ai_df = df[df.aio_content.notna()].copy().reset_index()
+        extractor = EntityExtractor(
+            model=entity_model,
+            records=ai_df,
+            attrs=aio_columns,
+            entities=ENTITIES,
+        )
+        ent_df = await extractor(n_concurrent=100)
 
-    return None
+        ent_df = ent_df[[col for col in ent_df.columns if col.startswith("entities_")]]
+        ent_df.columns = [f"aio_{col}" for col in ent_df.columns]
+        ent_df[id_column] = ai_df[id_column]
+        return ent_df
+    except Exception as exc:
+        LOG.error(f"Error processing AI overviews: {exc}")
+        return None
 
 
 def mentioned_in_string(
@@ -472,34 +556,15 @@ def add_brand_mentions(
     whole_word: bool = True,
 ) -> DataFrame:
     """Identify brand mentions in AI overviews using regex."""
-    if "aiOverview_content" in df.columns:
-        if brands:
-            df["aiOverview_brand_mentions"] = mentioned_brands(
-                df["aiOverview_content"],
-                brands=brands,
-                whole_word=whole_word,
-            )
-        if competitors:
-            df["aiOverview_competitor_mentions"] = mentioned_brands(
-                df["aiOverview_content"],
-                brands=competitors,
-                whole_word=whole_word,
-            )
-
-    if "aiOverview_source_titles" in df.columns:
-        if brands:
-            df["aiOverview_source_mentions"] = mentioned_brands(
-                df["aiOverview_source_titles"],
-                brands=brands,
-                whole_word=True,
-            )
-        if competitors:
-            df["aiOverview_source_competitor_mentions"] = mentioned_brands(
-                df["aiOverview_source_titles"],
-                brands=competitors,
-                whole_word=True,
-            )
-
+    for col in ("aio_content", "aio_sources"):
+        if col in df:
+            for tag, names in zip(("brand", "competitor"), (brands, competitors), strict=True):
+                if names:
+                    df[f"{col}_{tag}_mentions"] = mentioned_brands(
+                        df[col],
+                        brands=names,
+                        whole_word=whole_word,
+                    )
     return df
 
 
@@ -521,7 +586,11 @@ async def process_serps(
 
     df = features.merge(orgagg, on="term", how="left")
 
-    # Topics and intent
+    if cfg.brands or cfg.competitors:
+        LOG.info("Identifying brand and competitor mentions in AI overviews")
+        df = add_brand_mentions(df, brands=cfg.brands, competitors=cfg.competitors)
+
+    # AI topics and intent
     if cfg.topic_model is not None and cfg.assignment_model is not None:
         LOG.info("Extracting topics and intents from keywords and top organic results")
         topics = await topic_and_intent(
@@ -530,6 +599,8 @@ async def process_serps(
             topic_model=cfg.topic_model,
             assignment_model=cfg.assignment_model,
             max_retries=6,
+            text_column="term",
+            extra_columns=["titles", "descriptions"],
         )
         if topics is not None:
             df = df.merge(topics, on="term", how="left")
@@ -537,21 +608,17 @@ async def process_serps(
     # AI overview entities
     if cfg.entity_model is not None:
         LOG.info("Extracting entities from AI overviews")
-        entities = await process_ai_overviews(features, entity_model=cfg.entity_model)
+        entities = await extract_aio_entities(features, entity_model=cfg.entity_model)
         if entities is not None:
             df = df.merge(entities, on="term", how="left")
 
-    if cfg.brands or cfg.competitors:
-        LOG.info("Identifying brand and competitor mentions AI overviews")
-        df = add_brand_mentions(df, brands=cfg.brands, competitors=cfg.competitors)
-
     return df
 
 
 async def serps(cfg: SerpConfig, keywords: Iterable[str] | None = None) -> DataFrame | None:
     """Fetch and process SERP data for a list of keywords."""
     LOG.info(f"Fetching SERP data with config:\n{cfg}")
-    keywords = tuple(keywords) if keywords else None
+    keywords = tuple(keywords) if keywords is not None else None
     response = await fetch_serps(cfg=cfg, keywords=keywords)
 
     if response is None or len(response) == 0:
@@ -560,5 +627,5 @@ async def serps(cfg: SerpConfig, keywords: Iterable[str] | None = None) -> DataF
 
     LOG.info("Processing SERP data")
     df = await process_serps(response, cfg)
-    LOG.info(f"Got keyword dataframe:\n{df}")
+    LOG.info(f"Got SERPs dataframe:\n{df}")
     return df
